@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2011 Douglas Gilbert.
+ * Copyright (c) 2005-2013 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,12 +35,12 @@
 #define __STDC_FORMAT_MACROS 1
 #include <inttypes.h>
 
-#include "sdparm.h"
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
+#include "sdparm.h"
 
 /* sdparm_vpd.c : does mainly VPD page processing associated with the
- * INQUIRY SCSI command. Roughly in sync with spc4r21.
+ * INQUIRY SCSI command.
  */
 
 /* Prints outs an abridged set of device identification designators
@@ -184,6 +184,8 @@ decode_dev_ids_quiet(unsigned char * buff, int len, int m_assoc,
              */
             printf("%s\n", (const char *)ip);
             break;
+        case 9: /* PCIe routing ID */
+            break;
         default: /* reserved */
             break;
         }
@@ -455,6 +457,27 @@ decode_designation_descriptor(const unsigned char * ucp, int i_len,
          */
         printf("      %s\n", (const char *)ip);
         break;
+    case 9: /* Protocol specific port identifier */
+        /* added in spc4r36, PIV must be set, proto_id indicates */
+        /* whether UAS (USB) or SOP (PCIe) or ... */
+        if (! piv)
+            printf("      >>>> Protocol specific port identifier "
+                   "expects protocol\n"
+                   "           identifier to be valid and it is not\n");
+        if (TPROTO_UAS == p_id) {
+            printf("      USB device address: 0x%x\n", 0x7f & ip[0]);
+            printf("      USB interface number: 0x%x\n", ip[2]);
+        } else if (TPROTO_SOP == p_id) {
+            printf("      PCIe routing ID, bus number: 0x%x\n", ip[0]);
+            printf("          function number: 0x%x\n", ip[1]);
+            printf("          [or device number: 0x%x, function number: "
+                   "0x%x]\n", (0x1f & (ip[1] >> 3)), 0x7 & ip[1]);
+        } else
+            printf("      >>>> unexpected protocol indentifier: %s\n"
+                   "           with Protocol specific port "
+                   "identifier\n",
+                   sg_get_trans_proto_str(p_id, sizeof(b), b));
+        break;
     default: /* reserved */
         dStrHex((const char *)ip, i_len, 0);
         break;
@@ -612,8 +635,9 @@ decode_proto_lu_vpd(unsigned char * buff, int len)
 static int
 decode_proto_port_vpd(unsigned char * buff, int len)
 {
-    int k, bump, rel_port, desc_len, proto;
+    int k, j, bump, rel_port, desc_len, proto;
     unsigned char * ucp;
+    unsigned char * pidp;
 
     if (len < 4) {
         fprintf(stderr, "Protocol-specific port information VPD "
@@ -636,7 +660,13 @@ decode_proto_port_vpd(unsigned char * buff, int len)
         }
         if (desc_len > 0) {
             switch (proto) {
-            case TPROTO_SAS:
+            case TPROTO_SAS:    /* for SSP, added spl3r2 */
+                printf(" pwr_d_s=%d\n", !!(ucp[3] & 0x1));
+                pidp = ucp + 8;
+                for (j = 0; j < desc_len; j += 4, pidp += 4)
+                    printf("  phy id=%d, ssp persistent capable=%d\n",
+                           pidp[1], (0x1 & pidp[2]));
+                break;
             default:
                 fprintf(stderr, "Unexpected proto=%d\n", proto);
                 dStrHex((const char *)ucp, bump, 1);
@@ -773,6 +803,8 @@ decode_ext_inq_vpd(unsigned char * b, int len, int long_out, int protect)
         printf("  POA_SUP=%d\n", !!(b[12] & 0x80));     /* spc4r32 */
         printf("  HRA_SUP=%d\n", !!(b[12] & 0x40));     /* spc4r32 */
         printf("  VSA_SUP=%d\n", !!(b[12] & 0x20));     /* spc4r32 */
+        printf("  Maximum supported sense data length=%d\n",
+               b[13]); /* spc4r34 */
     } else {
         printf("  ACTIVATE_MICROCODE=%d SPT=%d GRD_CHK=%d APP_CHK=%d "
                "REF_CHK=%d\n", ((b[4] >> 6) & 0x3), ((b[4] >> 3) & 0x7),
@@ -792,6 +824,8 @@ decode_ext_inq_vpd(unsigned char * b, int len, int long_out, int protect)
                (b[10] << 8) + b[11]);
         printf("  POA_SUP=%d HRA_SUP=%d VSA_SUP=%d\n",      /* spc4r32 */
                !!(b[12] & 0x80), !!(b[12] & 0x40), !!(b[12] & 0x20));
+        printf("  Maximum supported sense data length=%d\n",
+               b[13]); /* spc4r34 */
     }
     return 0;
 }
@@ -956,8 +990,8 @@ decode_block_limits_vpd(unsigned char * buff, int len)
     if (len > 19) {     /* added in sbc3r09 */
         u = ((unsigned int)buff[16] << 24) | (buff[17] << 16) |
             (buff[18] << 8) | buff[19];
-        printf("  Maximum prefetch, xdread, xdwrite transfer length: %u "
-               "blocks\n", u);
+        printf("  Maximum prefetch length: %u blocks\n", u);
+        /* was 'Maximum prefetch transfer length' prior to sbc3r33 */
     }
     if (len > 27) {     /* added in sbc3r18 */
         u = ((unsigned int)buff[20] << 24) | (buff[21] << 16) |
@@ -1010,6 +1044,9 @@ decode_block_dev_chars_vpd(unsigned char * buff, int len)
     else
         printf("  Nominal rotation rate: %d rpm\n", u);
     u = buff[7] & 0xf;
+    printf("  Product type=%d\n", buff[6]);
+    printf("  WABEREQ=%d\n", (buff[7] >> 6) & 0x3);
+    printf("  WACEREQ=%d\n", (buff[7] >> 4) & 0x3);
     printf("  Nominal form factor");
     switch (u) {
     case 0:
@@ -1034,6 +1071,8 @@ decode_block_dev_chars_vpd(unsigned char * buff, int len)
         printf(": reserved\n");
         break;
     }
+    printf("  FUAB=%d\n", buff[8] & 0x2);
+    printf("  VBULS=%d\n", buff[8] & 0x1);
     return 0;
 }
 
@@ -1076,11 +1115,11 @@ decode_block_lb_prov_vpd(unsigned char * b, int len)
         return SG_LIB_CAT_MALFORMED;
     }
     printf("  Unmap command supported (LBPU): %d\n", !!(0x80 & b[5]));
-    printf("  Write same (16) with unmap bit supported (LBWS): %d\n", 
+    printf("  Write same (16) with unmap bit supported (LBWS): %d\n",
            !!(0x40 & b[5]));
     printf("  Write same (10) with unmap bit supported (LBWS10): %d\n",
            !!(0x20 & b[5]));
-    printf("  Logical block provisioning read zeros (LBPRZ): %d\n", 
+    printf("  Logical block provisioning read zeros (LBPRZ): %d\n",
            !!(0x4 & b[5]));
     printf("  Anchored LBAs supported (ANC_SUP): %d\n", !!(0x2 & b[5]));
     printf("  Threshold exponent: %d\n", b[4]);
@@ -1230,7 +1269,7 @@ decode_std_inq(int sg_fd, const struct sdparm_opt_coll * opts)
     printf("ACC=%d  TPGS=%d  3PC=%d  Protect=%d ",
            !!(b[5] & 0x40), ((b[5] & 0x30) >> 4), !!(b[5] & 0x08),
            !!(b[5] & 0x01));
-    printf(" BQue=%d\n  EncServ=%d  ", !!(b[6] & 0x80), !!(b[6] & 0x40));
+    printf(" [BQue=%d]\n  EncServ=%d  ", !!(b[6] & 0x80), !!(b[6] & 0x40));
     if (b[6] & 0x10)
         printf("MultiP=1 (VS=%d)  ", !!(b[6] & 0x20));
     else
@@ -1565,12 +1604,10 @@ sdp_process_vpd_page(int sg_fd, int pn, int spn,
             return 0;
         }
         up = b + 4;
-        for ( ; len > 5; len -= 6, up += 6) {
-            printf("    ");
-            for (k = 0; k < 6; ++k)
-                printf("%02x", (unsigned int)up[k]);
-            printf("\n");
-        }
+        for ( ; len > 5; len -= 6, up += 6)
+            printf("    IEEE Company_id: 0x%06x, vendor specific extension "
+                   "id: 0x%06x\n", (up[0] << 16) | (up[1] << 8) | up[2],
+                   (up[3] << 16) | (up[4] << 8) | up[5]);
         break;
     case VPD_UNIT_SERIAL_NUM:
         if (b[1] != pn)
@@ -1591,6 +1628,26 @@ sdp_process_vpd_page(int sg_fd, int pn, int spn,
         } else
             printf("  <empty>\n");
         break;
+    case VPD_3PARTY_COPY:
+        if (b[1] != pn)
+            goto dumb_inq;
+        len = (b[2] << 8) + b[3];       /* spc4r25 */
+        if (len > sz) {
+            fprintf(stderr, "Response to Third party copy VPD page "
+                    "truncated\n");
+            len = sz;
+        }
+        if (opts->long_out)
+            printf("Third party copy [0x8f] VPD page:\n");
+        else
+            printf("Third party copy VPD page:\n");
+        if (opts->hex) {
+            dStrHex((const char *)b, len + 4, 0);
+            return 0;
+        }
+        printf("    Leave decoding until needed, in hex:\n");
+        dStrHex((const char *)b, len + 4, 0);
+        break;
     case 0xb0:          /* VPD page depends on pdt */
         if (b[1] != pn)
             goto dumb_inq;
